#include <stdio.h>
#include <string.h>
#include "args.h"
#include "cf.h"
#include "dict.h"
#include "expr.h"
#include "global.h"
#include "talloc.h"
#include "type.h"

extern struct global global;
extern Atom *reverse ();	/* XXX in netfind.c */

/*******************************************************************
 * functions callable from the config file for argument processing *
 *******************************************************************/

/*
 * add 'a' to the list of constraints
 */

static Atom *
fn_addconstraint (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    global.constraints = append (global.constraints, cons (a, NULL));
    return integer (0);
}

/*
 * add 'a' to the list of directories
 */

static Atom *
fn_adddir (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    global.libpath = append (global.libpath, cons (a, NULL));
    return integer (0);
}

/*
 * return 1 if 'newlib' is already in 'list'
 */

static int
findlib (Atom *list, Atom *newlib)
{
    if (!listp (list) || !stringp (newlib) || !listp (this (list)))
	return 0;
    if (stringp (this (this (list))) &&
	strcmp (stringval (this (this (list))), stringval (newlib)) == 0)
	return 1;
    return findlib (next (list), newlib);
}

/*
 * add 'a' to the list of libraries if it's not there already
 */

static Atom *
fn_addlib (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    if (findlib (argstuff.libs, a))
	;
    else
	argstuff.libs =
	    append (argstuff.libs,
		    cons (cons (a, cons (cons (string (":mode"),
					       argstuff.mode),
					 NULL)),
			  NULL));
    return integer (0);
}

/*
 * return the Nth argument relative to the current position
 */

static Atom *
fn_arg (Atom *a)
{
    int offset;
    int i;

    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a))
	return error ();
    offset = intval (a);
    if (offset < 0)
	return string ("");
    for (i = 0; i < offset; ++i)
	if (argstuff.args[i] == NULL)
	    return string ("");
    return string (argstuff.args[offset]);
}

/*
 * bind2lib (libname [, nargs])
 *
 * associate the next 'nargs' args with 'libname'
 */

static Atom *
fn_bind2lib (Atom *a)
{
    int nargs;

    if (!listp (a))
	return error ();
    if (nullp (next (a)))
	nargs = 1;
    else if (integerp (next (a))) {
	nargs = intval (next (a));
	if (!nullp (next (next (a))))
	    return error ();
	if (nargs < 1)
	    return error ();
    }
    else
	return error ();
    argstuff.bound_libname = this (a);
    argstuff.bound_nargs = nargs;
    return fn_addlib (cons (this (a), NULL));
}


/*
 * set the link mode to the value of the string 'a'
 */

static Atom *
fn_setmode (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    if (argstuff.mode)
	unref (argstuff.mode);
    argstuff.mode = ref (a);
    return integer (0);
}

/*
 * set the verbose flag to the value represented by integer 'a'
 */

#define MAX(a,b) (((a) > (b)) ? (a) : (b))

static Atom *
fn_setverbose (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a))
	return error ();
    if (intval (a))
	global.verbosity = MAX (global.verbosity, intval (a));
    else
	global.verbosity = 0;
    return integer (0);
}

/*
 * (skip (n))
 *
 * skip the next 'n' arguments
 */

static Atom *
fn_skip (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a) || intval (a) < 0)
	return error ();
    argstuff.skip = intval (a);
    return integer (0);
}

/*
 * (substr string first [len])
 *
 * if len is missing, returns remainder of the string
 */

static Atom *
fn_substr (Atom *a)
{
    Atom *str;
    int first;
    int len;

    if (!listp (a) || !listp (next(a)))
	return error ();
    if (!stringp (this (a)))
	return error ();
    str = this (a);
    if (!integerp (this (next (a))))
	return error ();
    first = intval (this (next (a)));
    if (nullp (next (next (a))))
	len = strlen (stringval (str)) - first;
    else if (integerp (next (next (a))))
	len = intval (next (next (a)));
    else
	return error ();
    if (len < 0)
	return string ("");
    return stringn (stringval (str) + first, len);
}

/**********************************************************************/

static Atom *
lookup_id (char *a, void *context)
{
    return next (lookup (argstuff.dict, a));
}


/*
 * convert a char ** to a list of strings
 */

static Atom *
argv2list (char **argv, int argc)
{
    if (argc == 0)
	return NULL;
    if (argc == 1)
	return cons (string (argv[0]), NULL);
    return cons (string (argv[0]), argv2list (argv+1, argc-1));
}

/*
 * count the number of elements in a list
 *
 * XXX should move elsewhere?
 * XXX what's the normal LISP name for this function?
 */

static int
count (Atom *list)
{
    int result = 0;

    for (; !nullp (list); list = next (list))
	result++;
    return result;
}

/*
 * convert a list of strings to a char **
 *
 * NB: result vector is malloc'ed, individual strings are not
 */

char **
list2argv (Atom *list)
{
    char **result = TALLOC(count (list) + 1, char *);
    int i;

    for (i = 0; !nullp (list); list = next (list)) {
	if (stringp (this (list)))
	    result[i++] = stringval (this (list));
    }
    result[i] = NULL;
    return result;
}

/*
 * process command-line arguments according to the instructions in the
 * config file for this compiler
 */

void
process_args (char **argv, Atom *root)
{
    int i;
    Atom *newarg;
    Atom *config = lookup (root, ":args");

    argstuff.skip = 0;
    argstuff.libs = ref(global.required_libs);
    argstuff.mode = NULL;
    argstuff.dict = NULL;
    argstuff.args = NULL;
    argstuff.arglist = NULL;
    argstuff.bound_libname = NULL;
    argstuff.bound_nargs = 1;

    argstuff.dict = define (argstuff.dict, "addconstraint",
			    function (fn_addconstraint));
    argstuff.dict = define (argstuff.dict, "adddir", function (fn_adddir));
    argstuff.dict = define (argstuff.dict, "addlib", function (fn_addlib));
    argstuff.dict = define (argstuff.dict, "arg", function (fn_arg));
    argstuff.dict = define (argstuff.dict, "bind2lib", function (fn_bind2lib));
    argstuff.dict = define (argstuff.dict, "setmode", function (fn_setmode));
    argstuff.dict = define (argstuff.dict, "setverbose",
			    function (fn_setverbose));
    argstuff.dict = define (argstuff.dict, "skip", function (fn_skip));
    argstuff.dict = define (argstuff.dict, "substr", function (fn_substr));

    argstuff.mode = ref (string ("ds"));

    /*
     * if this compiler has the "Use-Compile-Args atrribute set to
     * "yes", attach an attribute to the compiler name
     *
     * XXX might be a good idea to associate all attributes with the
     * compiler name, that way we won't need to special-case them
     * here.
     */
    if (root &&
	(newarg = nlookup (root, "Use-Compile-Args")) &&
	strcmp (stringval (newarg), "yes") == 0) {
	newarg = cons (string (argv[0]),
		       cons (cons (string (":use-compile-args"),
				   integer (1)),
			     NULL));
    }
    else
	newarg = string (argv[0]);
    argstuff.arglist = append (argstuff.arglist, cons (newarg, NULL));

    for (i = 1; argv[i]; ++i) {
	Atom *pat = find_pattern (argv[i], config);

	argstuff.skip = 0;
	argstuff.args = argv + i;
	argstuff.bound_libname = NULL;
	argstuff.bound_nargs = 1;

	if (pat) {
#if 1
	    expr_evaluate (next (pat), lookup_id, NULL);
#else
	    printf ("argv[%d] = %s ", i, argv[i]);
	    fprint_atom (stdout, next (pat));
	    printf (" => ");
	    print_atom (expr_evaluate (next (pat), lookup_id));
	    printf ("\n");
	    
	    printf ("skip = %d\n", argstuff.skip);
	    printf ("libs = ");
	    print_atom (argstuff.libs);
	    printf ("\n");
	    printf ("mode = %s\n", stringval (argstuff.mode));
	    printf ("verbose = %d\n", argstuff.verbose);
#endif
	}

	if (argstuff.bound_libname) {
	    newarg = cons (argv2list (argstuff.args, argstuff.bound_nargs),
			   cons (cons (string (":lib"),
				       argstuff.bound_libname),
				 NULL));
	}
	else {
	    newarg = string (argv[i]);
	}
	argstuff.arglist = append (argstuff.arglist, cons (newarg, NULL));
	while (argstuff.skip > 0 && argv[i+1] != NULL) {
	    argstuff.arglist = append (argstuff.arglist,
				       cons (string (argv[++i]), NULL));
	    --argstuff.skip;
	}
    }
}

/*
 * convert a PATH-style string (directories delimited by ":")
 * to a list of strings
 */

Atom *
path2list (char *s)
{
    char *colon;

    if (s == NULL)
	return NULL;
    else if (*s == '\0')
	return cons (string (""), NULL);
    else if (colon = strchr (s, ':')) 
	return cons (stringn (s, colon-s), path2list (colon+1));
    else
	return cons (string (s), NULL);
}

/*
 * convert a list to a PATH-style string
 */

char *
list2path (Atom *list)
{
    Atom *ptr;
    int length;
    char *result;

    length = 0;
    for (ptr = list; !nullp(ptr); ptr = next (ptr))
	length += strlen (stringval (this (ptr))) + 1;
    result = xalloc (length);
    for (ptr = list; !nullp(ptr); ptr = next (ptr)) {
	if (ptr == list)
	    strcpy (result, stringval (this (ptr)));
	else {
	    strcat (result, ":");
	    strcat (result, stringval (this (ptr)));
	}
    }
    return result;
}

/*
 * given two lists l1 and l2, return the concatenation of the two
 * (similar to append, but doesn't modify l1)
 *
 * XXX I'm showing my ignorance here.  there's probably a standard
 * LISP function that does this.
 */

static Atom *
collapse (Atom *l1, Atom *l2)
{
    if (nullp (l1))
	return l2;
    else if (listp (l1))
	return cons (this (l1), collapse (next (l1), l2));
    else
	return cons (l1, l2);
}

/*
 * compute the arguments to be used to link this library
 * from the library.linkarg# attributes in the metadata
 */

static Atom *
extract_lib_args (Atom *lib, Atom *metadata)
{
    int i;
    char buf[200];
    Atom *argproto;
    Atom *result = NULL;
    char *localdir = stringval (nlookup (metadata, "localdir"));

    for (i = 0; ;++i) {
	char *arg;

	sprintf (buf, "library.linkarg%d", i);
	if ((argproto = nlookup (metadata, buf)) == NULL)
	    break;
	arg = TALLOC (strlen (stringval(argproto)) + 4 * strlen (localdir) + 1,
		      char);
	sprintf (arg, stringval (argproto),
		 localdir, localdir, localdir, localdir);
	result = cons (string (arg), result);
	free (arg);
    }
    return reverse (result);
}

/*
 * return the default arguments for linking library 'libname'
 * from directory "localdir" (in metadata)
 */

static Atom *
default_args (Atom *libname, Atom *metadata)
{
    Atom *localdir = nlookup (metadata, "localdir");

    if (localdir)
	return cons (string_concat (string ("-L"), localdir, NULL),
		     string_concat (string ("-l"), libname, NULL));
    else
	return string_concat (string ("-l"), libname, NULL);
}

/*
 * figure out what args to use to link 'libname'.
 *
 * if 'libname' is in 'liblist',
 *     if the metadata for 'libname' specifies args
 *         use the args specified in the metadata;
 *     else
 *         use default args
 * else
 *     use original_args
 */
 
static Atom *
get_lib_args (Atom *libname, Atom *liblist, Atom *original_args)
{
    Atom *lib = nlookup (liblist, stringval (libname));

    if (lib) {
	Atom *metadata = unmeta ("candidate", next (lib));

	if (metadata) {
	    if (nlookup (metadata, "library.linkarg0") != NULL)
		return extract_lib_args (lib, metadata);
	    else
		return default_args (libname, metadata);
	}
	else {
	    return string ("bogus_metadata"); /* XXX */
	}
    }
    else {
	/* lib not found in list */
	return original_args;
    }
}

/*
 * generate the arguments that needed to be added to a compile command
 * for a source file that references the lib described in metadata
 */

static Atom *
get_compile_args (Atom *metadata)
{
    Atom *result = NULL;
    char *localdir;
    char buf[200];
    int i;

    if (metadata == NULL)
	return NULL;

    localdir = stringval (nlookup (metadata, "localdir"));	

    for (i = 0; ; ++i) {
	char *arg;
	Atom *argproto;

	sprintf (buf, "library.compilearg%d", i);
	if ((argproto = nlookup (metadata, buf)) == NULL)
	    break;

	if (localdir) {
	    arg = TALLOC (strlen (stringval(argproto)) + 4 *
			  strlen (localdir) + 1,
			  char);
	    sprintf (arg, stringval (argproto),
		     localdir, localdir, localdir, localdir);
	    result = cons (string (arg), result);
	    free (arg);
	}
	else {
	    result = cons (argproto, result);
	}
    }
    return reverse (result);
}

/*
 * given a an augmented list of command-line arguments ('arglist')
 * and a list of library-to-metadata bindings ('liblist')
 * generate an argument list for a child process to build this executable
 *
 * make_child_arglist() is called from nbshim.c and builds the entire
 * command-line, including adding any args that aren't associated with
 * the original command-line.
 *
 * make_child_arglist_r() handles the rest of the command line arguments
 */

static Atom *
make_child_arglist_r (Atom *arglist, Atom *liblist)
{
    Atom *rest;

    if (nullp (arglist))
	return NULL;
    rest = make_child_arglist_r (next (arglist), liblist);
    if (stringp (this (arglist)))
	return cons (this (arglist), rest);
    else if (listp (this (arglist))) {
	Atom *lib;

	if (lib = nlookup (next (this (arglist)), ":lib"))
	    return collapse (get_lib_args (lib, liblist, this (this (arglist))), rest);
	else
	    return collapse (this (this (arglist)), rest);
    }
}

Atom *
make_child_arglist (Atom *arglist, Atom *liblist)
{
    Atom *compile_args = NULL;
    Atom *ptr;
    Atom *rest;
    
    rest = make_child_arglist_r (next (arglist), liblist);

    if (listp (this (arglist))) {
	Atom *arg0 = this (this (arglist));

	if (intval (nlookup (next (this (arglist)),
			     ":use-compile-args")) == 1) {
	    for (ptr = liblist; !nullp (ptr); ptr=next (ptr)) {
		Atom *lib = this (ptr);
		Atom *metadata = unmeta ("candidate", next (next (lib)));
		
		if (metadata) {
		    if (nlookup (metadata, "library.compilearg0") != NULL)
			compile_args = append (compile_args,
					       get_compile_args (metadata));
		}
	    }
	}
	return cons (arg0, collapse (compile_args, rest));
    }
    else
	return cons (this (arglist), rest);
}


#ifdef TEST
struct global global;

main (int argc, char **argv)
{
    Atom *root = find_config ("default.config", argv[1]);
    Atom *argpats = lookup (root, ":args");
    Atom *path;
    int i;

#if 0
    fprint_atom (stdout, root);
    printf ("\n");
#endif

    if (path = lookup (root, "Library-Path"))
	global.libpath = path2list (stringval (next (path)));

    process_args (argv+1, root);

    printf ("libs: ");
    print_atom (argstuff.libs);
    printf ("\narglist:");
    print_atom (argstuff.arglist);
    printf ("\nchild arglist:");
    print_atom (make_child_arglist (argstuff.arglist, NULL));
    printf ("\n");
}
#endif
